Stream = require('stream')
Zlib = require('zlib')

Request = require('request')
SocketIoStream = require('socket.io-stream')

Base = require('../base')
Logger = require('./logger')


debug = Base.logger('http_proxy')

PROXY_HTTP_HOST_USE_PROXY = 'USE_PROXY'

upstreamUrl = null
isUpstreamHttps = null

# Rewriting vars.
isRewriteHostOn = null
upstreamHost = null
upstreamProto = null
upstreamHostReStr = null
upstreamHostRe = null
rewriteProtoDelimReStr = ":\\\\?\\/\\\\?\\/" # supports '://'  or ':\/\/'
rewriteCustomReplacers = null
rewriteHeadersCustomReplacers = null
rewriteUseRealDeflate = null


exports.setupHttpProxy = (upstream) ->
  upstreamUrl = upstream || Base.env.IX_HTTP_PROXY
  upstreamUrl = upstreamUrl?.trim().replace(/\/$/, '')
  throw new Error("No upstream server defined.") unless upstreamUrl
  debug('setupHttpProxy upstreamUrl', upstreamUrl)
  isUpstreamHttps = upstreamUrl.slice(0,6) == 'https:'
  isRewriteHostOn = Base.env.IX_PROXY_REWRITE_HOST?.toLowerCase() == 'on'
  if isRewriteHostOn
    upstreamHost = upstreamUrl.replace(/^https?:\/\//i,'')
    upstreamProto = if isUpstreamHttps then 'https' else 'http'
    rewriteUseRealDeflate = Base.env.IX_PROXY_REWRITE_REAL_DEFLATE?.toLowerCase() == 'on'
    upstreamHostReStr = regexEscape(upstreamHost)
    upstreamHostRe = new RegExp("\\b#{upstreamHostReStr}\\b", "ig")
    if Base.env.IX_PROXY_REWRITE_PROTO_DELIM_REGEX
      rewriteProtoDelimReStr += "|"+Base.env.IX_PROXY_REWRITE_PROTO_DELIM_REGEX
    customRes = Base.env.IX_PROXY_REWRITE_CUSTOM_REGEXES
    if customRes
      console.log("Parsing IX_PROXY_REWRITE_CUSTOM_REGEXES: #{customRes}")
      replacers = JSON.parse(customRes)
      rewriteCustomReplacers = []
      for replacer in replacers
        flags = "g"
        flags += replacer[2] if replacer.length > 2
        rewriteCustomReplacers.push([new RegExp(replacer[0], flags), replacer[1]])
      debug('rewriteCustomReplacers', rewriteCustomReplacers)
    customHeaderRes = Base.env.IX_PROXY_REWRITE_HEADERS_REGEXES
    if customHeaderRes
      console.log("Parsing IX_PROXY_REWRITE_HEADERS_REGEXES: #{customHeaderRes}")
      replacers = JSON.parse(customHeaderRes)
      rewriteHeadersCustomReplacers = []
      for replacer in replacers
        # [ nameRe, search, replace [, flags] ]
        flags = "g"
        flags += replacer[3] if replacer.length > 3
        rewriteHeadersCustomReplacers.push([new RegExp(replacer[0]), new RegExp(replacer[1], flags), replacer[2]])
      debug('rewriteHeadersCustomReplacers', rewriteHeadersCustomReplacers)

exports.onHttpProxy = (faasReq, raw_cb_proxyResponse) ->
  debug('onHttpProxy.faasReq.headers', faasReq?.headers)
  faasReq.origHost = faasReq.headers.host
  faasReq.origProto = faasReq.headers['x-forwarded-proto']
  Logger.logHttpRequest(faasReq)
  cb_proxyResponse = (pr) ->
    debug('cb_proxyResponse', { headers: pr?.headers, lag: pr?.lag, statusCode: pr?.statusCode, statusMessage: pr?.statusMessage, error: pr?.error })
    # debug('cp_proxyResponse.responseStream', pr?.responseStream)
    raw_cb_proxyResponse(pr)
  url = upstreamUrl + faasReq.url
  fwdOptions =
    url: url
    headers: faasReq.headers
    method: faasReq.method
    followRedirect: false
  httpHost = Base.env.IX_PROXY_HTTP_HOST
  if isUpstreamHttps || httpHost == PROXY_HTTP_HOST_USE_PROXY
    delete fwdOptions.headers.host
  if httpHost && httpHost != PROXY_HTTP_HOST_USE_PROXY
    fwdOptions.headers.host = httpHost
  if isRewriteHostOn
    delete fwdOptions.headers.host
    delete fwdOptions.headers['x-forwarded-host']
    delete fwdOptions.headers['x-forwarded-server'] # Some apps might use this (Wix)
  debug('request', fwdOptions)
  fwdReq = null
  requestStream = faasReq.requestStream
  proxyResponse = {}
  reqCounter = new Stream.Transform()
  reqCounter.numBytes = 0
  reqCounter._transform = (chunk, encoding, callback) ->
    debug('request chunk', encoding, chunk)
    console.log(chunk.toString())
    reqCounter.numBytes += chunk.length
    callback(null, chunk)
  try
    fwdReq = Request(fwdOptions)
  catch e
    # Error: Invalid protocol: localhost:
    proxyResponse.statusCode = 500
    proxyResponse.error = e.toString()
    Logger.logHttpResponseError(faasReq, proxyResponse)
    cb_proxyResponse(proxyResponse)
    requestStream.on('data', (data) -> debug('fwdReq.error data', data) )
    # requestStream.end(e.toString())
    return
  responseStarted = false
  responseStream = SocketIoStream.createStream()
  proxyResponse.responseStream = responseStream
  faasReq.responseBytes = 0
  respCounter = new Stream.Transform()
  respCounter._transform = (chunk, encoding, callback) ->
    debug('response chunk', encoding, chunk)
    faasReq.responseBytes += chunk.length
    callback(null, chunk)
  fwdReq.on 'error', (err) ->
    debug('fwdReq.error', err, proxyResponse)
    if responseStarted
      console.error('duplicate ERROR response attempted for '+url)
      responseStream.emit('error', err)
    else
      responseStarted = true
      proxyResponse.error = err.toString()
      cb_proxyResponse(proxyResponse)
    try
      fwdReq.abort()
      fwdReq.destroy()
      requestStream.unpipe()
      requestStream.on('data', (data) -> debug('fwdReq.error data', data) )
      responseStream.end()
    catch e
      console.error(e)
    Logger.logHttpResponseError(faasReq, proxyResponse)

  fwdReq.on 'response', (response) ->
    Logger.logHttpResponseStart(faasReq, response)
    debug('fwdReq.response', response.headers, response.statusCode, response.statusMessage)
    # debug('fwdReq.response.proxyResponse', proxyResponse)
    if responseStarted
      console.error('duplicate SUCCESS response attempted for '+url)
      return
    responseStarted = true
    proxyResponse.lag = faasReq.ts.responseStart - faasReq.ts.requestStart
    proxyResponse.headers = response.headers
    proxyResponse.statusCode = response.statusCode
    proxyResponse.statusMessage = response.statusMessage
    stream = fwdReq
    if isRewriteHostOn
      # Rewrite 301/302 redirects
      rewriteHeaders(proxyResponse.headers, faasReq.origHost, faasReq.origProto)
      ce = proxyResponse.headers['content-encoding']?.toLowerCase()
      isGzip = (ce == 'gzip' || ce == 'x-gzip')
      isDeflate = (ce == 'deflate')
      ct = proxyResponse.headers['content-type']?.toLowerCase() || ''
      canRewrite = (ct.indexOf('text/') > -1 || ct.indexOf('javascript') > -1 || ct.indexOf('js') > -1)
      canRewrite &&= (!ce || ce == 'identity' || isGzip || isDeflate)
      if canRewrite
        # We may change the size during rewrite,
        # so kill content-length and accept-ranges
        # (as the Range request would be forwarded),
        # and send it as chunked (opposite of content-length).
        delete proxyResponse.headers['content-length']
        proxyResponse.headers['transfer-encoding'] = 'chunked'
        delete proxyResponse.headers['accept-ranges']
        if isGzip
          stream = safePipe(stream, Zlib.createGunzip(), 'Gunzip')
        else if isDeflate
          # 'deflate' was screwed up by Microsoft.
          # Assume it's raw. Shouldn't be according to spec,
          # but we operate in the real world.
          # http://stackoverflow.com/questions/883841/why-do-real-world-servers-prefer-gzip-over-deflate-encoding
          if rewriteUseRealDeflate
            stream = safePipe(stream, Zlib.createInflate(), 'Inflate')
          else
            stream = safePipe(stream, Zlib.createInflateRaw(), 'InflateRaw')
        stream = safePipe(stream, newRewriterStream(faasReq.origHost, faasReq.origProto), 'Rewriter')
        if isGzip
          stream = safePipe(stream, Zlib.createGzip(), 'Gzip')
        else if isDeflate
          if rewriteUseRealDeflate
            stream = safePipe(stream, Zlib.createDeflate(), 'Deflate')
          else
            stream = safePipe(stream, Zlib.createDeflateRaw(), 'DeflateRaw')
    cb_proxyResponse(proxyResponse)
    ###
    # TODO: Better handling of errors!
    # To test, manually force a wrong decoder (e.g. "if isGzip || true").
    # Need to incorporate fwdReq.on('error',...) logic at any point in chain.
    ###
    if stream == fwdReq
      stream.pipe(respCounter).pipe(responseStream)
    else
      safePipe(safePipe(stream, respCounter, 'respCounter'), responseStream, 'responseStream')

  fwdReq.on 'end', ->
    Logger.logHttpResponseEnd(faasReq)

  requestStream.pipe(reqCounter).pipe(fwdReq)

newRewriterStream = (host, proto) ->
  queue = ''
  if upstreamProto != proto
    hostsReStr = regexEscape(host)
    if Base.env.IX_PROXY_REWRITE_PROTO_HOSTS_REGEX
      hostsReStr += "|"+Base.env.IX_PROXY_REWRITE_PROTO_HOSTS_REGEX
    changeProtoOriginsRe = new RegExp("\\b#{upstreamProto}(#{rewriteProtoDelimReStr})(#{hostsReStr})\\b", "ig")
    debug('changeProtoOriginsRe', changeProtoOriginsRe)
  rewriter = new Stream.Transform()
  rewriter._transform = (chunk, encoding, callback) ->
    debug('rewriter chunk', encoding, chunk)
    queue += chunk.toString('utf8')
    queue = queue.replace(upstreamHostRe, host)
    if upstreamProto != proto
      queue = queue.replace(changeProtoOriginsRe, "#{proto}$1$2")
    if rewriteCustomReplacers
      for replacer in rewriteCustomReplacers
        queue = queue.replace(replacer[0], replacer[1])
    if queue.length > 100
      splitPoint = queue.length - 100
      oldChunk = queue.slice(0, splitPoint)
      queue = queue.slice(splitPoint)
      rewriter.push(new Buffer(oldChunk, 'utf8'))
    callback()
  rewriter._flush = (callback) ->
    debug('rewriter flush')
    rewriter.push(new Buffer(queue, 'utf8'))
    callback()
  rewriter

regexEscape = (str) ->
  # http://stackoverflow.com/questions/2593637/how-to-escape-regular-expression-in-javascript/17326679#17326679
  str.replace(/([.*+?^=!:${}()|\[\]\/\\])/g, "\\$1")

rewriteHeaders = (headers, host, proto) ->
  # Always rewrite Location: (for 301/302 redirects).
  if headers.location
    headers.location = headers.location.replace(upstreamHostRe, host)
    if upstreamProto != proto
      hostReStr = regexEscape(host)
      headers.location = headers.location.replace(new RegExp("\\b#{upstreamProto}(#{rewriteProtoDelimReStr})(#{hostReStr})\\b", "ig"), "#{proto}$1$2")
  # Always rewrite Set-Cookie: domain=...
  if headers['set-cookie']
    domainRe = new RegExp("(;\\s+domain=\\s*\\.?)#{upstreamHostReStr}(\\s|;|$)", "ig")
    for val, i in headers['set-cookie']
      headers['set-cookie'][i] = val.replace(domainRe, "$1#{host}$2")
  # Possibly rewrite other things - higher-up Set-Cookie domains, etc.
  if rewriteHeadersCustomReplacers
    for replacer in rewriteHeadersCustomReplacers
      nameRe = replacer[0]
      search = replacer[1]
      replace = replacer[2]
      for own name, value of headers
        if name.match(nameRe)
          if Array.isArray(value)
            for val, i in value
              value[i] = val.replace(search, replace)
          else
            headers[name] = value.replace(search, replace)

safePipe = (inpipe, outpipe, lbl) ->
  outpipe.label = lbl
  label = "stream.#{inpipe.label}"
  inpipe.pipe(outpipe)
  inpipe.on 'error', (err) ->
    console.error(["#{label}.error", err])
    ###
    # responseStream (SocketIoStream) doesn't seem to respond well
    # to ".emit('error',...)". So .end() it instead of propogating error.
    ###
    # try outpipe.emit('error', err) catch e then debug("#{label}.outpipe.emitError.error", e)
    try outpipe.end() catch e then debug("#{label}.end.error", e)
    try inpipe.unpipe(outpipe) catch e then debug("#{label}.unpipe.error", e)
    try inpipe.on('data', -> debug("#{label}.data", data)) catch e then debug("#{label}.data.error", e)
  inpipe.on('end', -> debug("#{label}.end"))
  outpipe
